#ifndef DECODER_H
#define DECODER_H
#include "..\Component.h"
#include "..\Components\CheckNodeFunctionalUnit.h"
#include "..\Components\SymbolNodeFunctionalUnit.h"
#include "..\Components\Multiplexer.h"
#include "..\Components\BasicComponent.h"
#include "..\Scheduling\FinitStateMachine.h"
#include "..\Scheduling\DecodeScheduling.h"
#include "..\Components\VariableShiftRegister.h"
#include "..\Components\Memory.h"
#include "..\Components\PermutationNetwork.h"
#include "..\..\..\ControlInformation\widths.h"
#include "..\Scheduling\ControlCN.h"

class Decoder : public Component
{
public:
#pragma region Interface
	// input side
	sc_in<TYPE > in_received_value;					// received SD-Values
	sc_in<sc_logic> in_new_values_present;				// this flag determines wether the next received code word is available
	sc_out<sc_logic> out_ready_for_next_received_value;	// this flag indicates, that the decoder ist ready for the next received symbol
	sc_out<sc_lv<WIDTH_ITERATION> > out_iterations;

	// output side
	sc_out<sc_logic> out_decoded_value;					// decoded HD-Values
	sc_out<sc_logic> out_decoded_value_present;
	sc_out<sc_logic> out_decoding_valid;				// indicates success of the decoding
#pragma endregion
	sc_signal<TYPE > received_value;

	// Symbol Node Functional Unit
	SymbolNodeFunctionalUnit** snFU;
	sc_signal<TYPE > sn_soft_decision[PARALLELISM];
	sc_signal<TYPE > sn_extrinsic_message[PARALLELISM];
	sc_signal<sc_logic> sn_init[PARALLELISM];
	sc_signal<sc_logic> sn_output[PARALLELISM];
	sc_signal<sc_lv<PARALLELISM> > sn_init_vector;
	sc_signal<sc_lv<PARALLELISM> > sn_output_vector;
	sc_signal<sc_logic> oddIteration;
	sc_signal<sc_logic> sn_readEnable;
	sc_signal<bool> evenIteration;
	sc_signal<sc_logic> sn_writeEnable;
	sc_signal<sc_lv<DEGREE_WIDTH> > sn_variableDelay;
	Multiplexer<TYPE >** mux1; // Multiplexer on the input of SN - needed to switch between initialization with received values and decoding
	
	// Check Node Functional Unit
	CheckNodeFunctionalUnit** cnFU;
	sc_signal<TYPE > cn_soft_decision[PARALLELISM];
	sc_signal<TYPE > cn_extrinsic_message[PARALLELISM];
	sc_signal<sc_logic> cn_valid[PARALLELISM];
	sc_signal<sc_lv<PARALLELISM> > cns_valid;
	sc_signal<sc_logic> cn_first_value;
	sc_signal<sc_logic> cn_last_value;
	sc_signal<sc_lv<DEGREE_WIDTH> >	cn_index_in;
	sc_signal<sc_lv<DEGREE_WIDTH> >	cn_index_out;
	sc_signal<bool> cn_first_iteration;
	sc_signal<sc_logic> cn_MEM_enable;
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > cn_mem_address;
	sc_signal<sc_lv<DEGREE_WIDTH> > cn_variableDelay;
	VariableShiftRegister<sc_logic>* variableShift3;

	// Finite State Machine
	FiniteStateMachine* fsm;
	sc_signal<sc_lv<ROM_SN_WIDTH-2> > sn_address;
	sc_signal<sc_logic> fsm_enable_decode_scheduling;
	sc_signal<sc_logic> fsm_reset_ds;
	sc_signal<sc_lv<WIDTH_ITERATION> > iteration;

	// Decode Scheduling
	DecodeScheduling *decodeScheduling;
	sc_signal<sc_lv<ROM_ADDRESS_WIDTH> > ds_control_MEM_address;
	sc_signal<sc_logic> ds_MEM_enable;
	sc_signal<sc_lv<DEGREE_WIDTH> > ds_cn_degree;
	sc_signal<sc_lv<DEGREE_WIDTH> > ds_cn_index;
	sc_signal<sc_logic> ds_cn_first_value;
	sc_signal<sc_logic> ds_cn_first_value_delayed;
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > ds_cn_mem_address;
	sc_signal<sc_logic> write_to_output_enable[PARALLELISM];

	SC_HAS_PROCESS(Decoder);
	Decoder()
	{
		SC_METHOD(Iteration);
		sensitive << this->iteration;
	}

	void Iteration()
	{
		// Odd / Even Iteration for multiplexing between two sd-memory banks and first/last values;
		sc_lv<WIDTH_ITERATION> iteration = this->iteration.read();
		this->oddIteration.write(iteration[0]);
		if(iteration.or_reduce())
		{
			this->cn_first_iteration.write(false);
			for(int i = 0; i < PARALLELISM; i++)
			{
#ifndef MTI_SYSTEMC
				//this->snFU[i]->Print();
				//this->cnFU[i]->Print();
#endif //MTI_SYSTEMC
			}
		}
		else
		{
			this->cn_first_iteration.write(true);
		}
	}
};
#endif // SERIALDECODER_H
