#ifndef PARALLELOPMP_H
#define PARALLELOPMP_H
#include "Decoder.h"
#include "..\Scheduling\ControlSNandPI.h"
#include "..\Scheduling\InitAndOutput.h"

#ifndef MTI_SYSTEMC
class ParallelOPMP: public Decoder
{
public:
	
	// Validation
	sc_signal<sc_logic> decoding_valid;
	sc_signal<sc_logic> lastIterationReached;

	// Permutation Networks
	PermutationNetwork** permutation_networks;
	sc_signal<sc_lv<ROM_PERMUTATION_WIDTH> > permutation_networks_indices[4];
	sc_signal<sc_lv<PARALLELISM * WIDTH_SD> > pi_input[4];
	sc_signal<sc_lv<PARALLELISM * WIDTH_SD> > pi_output[4];


	sc_signal<TYPE > soft_decision_permutations[4 * PARALLELISM];
	sc_signal<TYPE > cn_extrinsic_message_permuted[PARALLELISM];

	sc_signal<TYPE > sn_soft_decision_delayed[PARALLELISM];
	sc_signal<TYPE > sn_output_permuted[PARALLELISM];
	Multiplexer<TYPE >** mux2;
	
	ShiftRegister<TYPE >** delay_of_SN_output;
	ShiftRegister<TYPE >* sd_input;

	sc_signal<sc_logic> ds_cn_first_value_out;

	// InitWriteBack
	InitAndOutput*					initAndOutput;
	sc_signal<sc_logic>				initState;
	sc_signal<sc_logic>				writeBackState;
	sc_signal<sc_logic>				loadAndwriteBackDone;

	// Check Node Control
	ControlCN*						controlCN;
	sc_signal<sc_logic>				cn_mem_read;
	sc_signal<sc_logic>				cn_last_value_delayed;
	
	// Symbol Node Memory Access
	ControlSNandPI* controlSNandPI;
	sc_signal<sc_lv<ROM_SN_WIDTH - 2> > oldSD_address[PARALLELISM];
	sc_signal<sc_lv<ROM_SN_WIDTH - 2> > newSD_address[PARALLELISM];
	sc_signal<sc_logic>					first_extrinsic[PARALLELISM];
	sc_signal<sc_logic>					last_extrinsic[PARALLELISM];
	sc_signal<sc_logic>					newSD_readEnable;
	sc_signal<sc_logic>					oldSD_readEnable;
	sc_signal<sc_lv<ROM_SN_WIDTH * PARALLELISM> > romSN_output;
	sc_signal<sc_lv<(ROM_SN_WIDTH - 2) * PARALLELISM> >	oldSD_address_vector;
	sc_signal<sc_lv<4 * ROM_PERMUTATION_WIDTH> > permutation_networks_indices_vector;
	

	ShiftRegister<sc_logic >** delay_write_to_output;

	SC_HAS_PROCESS(ParallelOPMP);
	ParallelOPMP(sc_module_name name, const char* degreeFileName, int max_iterations)
	{
		/* *************************** Combinatorical Processes  *************************** */
		SC_METHOD(SNAddress);
		sensitive << sn_address;
		for(int i = 0; i < PARALLELISM; i++)
		{
			sensitive << oldSD_address_vector << this->sn_init[i] << this->sn_output[i];
		}
		SC_METHOD(Validity);
		for(int i = 0; i < PARALLELISM; i++)
		{
			sensitive << this->cn_valid[i];
		}

		SC_METHOD(SNInitOutputControl);
		sensitive << this->sn_init_vector << this->sn_output_vector;

		SC_METHOD(RomSNOutput);
		sensitive << this->romSN_output;

		SC_METHOD(PermutationOutput);
		for(int i = 0; i < 4; i++)
		{
			sensitive << this->pi_output[i];
		}
		SC_METHOD(PermutationInput);
		for(int j = 0; j < PARALLELISM; j++)
		{
			sensitive << sn_output_permuted[j] << cn_extrinsic_message[j] << soft_decision_permutations[j * 4] << soft_decision_permutations[j * 4 + 2];
		}
		SC_METHOD(PermutationOutputMapping);
		sensitive << this->permutation_networks_indices_vector;

		/* ************************ Initialization and Write Back  ************************* */
		this->initAndOutput = new InitAndOutput("InitWriteBack");
		this->initAndOutput->BindCommonPorts(this);
		this->initAndOutput->in_init(this->initState);
		this->initAndOutput->in_writeBack(this->writeBackState);
		this->initAndOutput->out_loadAndwriteBackDone(this->loadAndwriteBackDone);
		this->initAndOutput->out_address(sn_address);
		this->initAndOutput->out_init(this->sn_init_vector);
		this->initAndOutput->out_writeBack(this->sn_output_vector);

		/* ********************** SN and Permutation Network Control  ********************** */
		this->controlSNandPI = new ControlSNandPI("AccessRom");
		this->controlSNandPI->BindCommonPorts(this);
		this->controlSNandPI->in_enableRom(this->ds_MEM_enable);
		this->controlSNandPI->in_addressRom(this->ds_control_MEM_address);
		this->controlSNandPI->in_variableDelay(this->sn_variableDelay);
		this->controlSNandPI->in_oddIteration(this->oddIteration);

		this->controlSNandPI->out_newSD_readEnable(this->newSD_readEnable);
		this->controlSNandPI->out_oldSD_readEnable(this->oldSD_readEnable);
		this->controlSNandPI->out_newSD_address_vector(this->romSN_output);
		this->controlSNandPI->out_oldSD_address_vector(this->oldSD_address_vector);
		this->controlSNandPI->out_permutationIndices_vector(this->permutation_networks_indices_vector);
		

		/* ****************************** Finite State Machine ***************************** */
		this->fsm = new FiniteStateMachine("FSM");
		this->fsm->BindCommonPorts(this);
		this->fsm->in_new_values_present(this->in_new_values_present);
		this->fsm->out_ready_for_next_received_value(this->out_ready_for_next_received_value);
		this->fsm->in_decoding_valid(this->decoding_valid);
		this->fsm->in_current_iteration(this->iteration);
		this->fsm->out_enable_decode_schedulig(this->fsm_enable_decode_scheduling);
		this->fsm->out_reset_decode_scheduling(this->fsm_reset_ds);
		this->fsm->out_decoding_valid(this->out_decoding_valid);
		this->fsm->out_iterations(this->out_iterations);
		this->fsm->in_last_iteration_reached(this->lastIterationReached);
		this->fsm->out_init(this->initState);
		this->fsm->out_writeBack(this->writeBackState);
		this->fsm->in_loadAndWriteBackDone(this->loadAndwriteBackDone);

		/* ******************************* Decode Scheduling ****************************** */
		this->decodeScheduling = new DecodeScheduling("DecodeScheduling", degreeFileName, max_iterations);
		this->decodeScheduling->clock(this->clock);
		this->decodeScheduling->enable(this->fsm_enable_decode_scheduling);
		this->decodeScheduling->reset(this->fsm_reset_ds);
		this->decodeScheduling->out_first_value(this->ds_cn_first_value_out);
		this->decodeScheduling->out_rom_enable(this->ds_MEM_enable);
		this->decodeScheduling->out_rom_index(this->ds_control_MEM_address);
		this->decodeScheduling->out_cn_degree(this->ds_cn_degree);
		this->decodeScheduling->out_cn_index(this->ds_cn_index);
		this->decodeScheduling->out_iteration(this->iteration);
		this->decodeScheduling->out_last_iteration_reached(this->lastIterationReached);
		
		/* ***************************** Symbol Node Functional Unit **************************** */
		this->snFU = (SymbolNodeFunctionalUnit**) malloc(sizeof(SymbolNodeFunctionalUnit*) * PARALLELISM);
		this->mux1 = (Multiplexer<TYPE >**) malloc(sizeof(Multiplexer<TYPE >*) * PARALLELISM);
		this->mux2 = (Multiplexer<TYPE >**) malloc(sizeof(Multiplexer<TYPE >*) * PARALLELISM);
		this->delay_of_SN_output = (ShiftRegister<TYPE >**) malloc(sizeof(ShiftRegister<TYPE >*) * PARALLELISM);
		this->delay_write_to_output = (ShiftRegister<sc_logic >**) malloc(sizeof(ShiftRegister<sc_logic >*) * PARALLELISM);
		for(int i = 0; i < PARALLELISM; i++)
		{
			this->snFU[i] = new SymbolNodeFunctionalUnit(PosIntToString("SnFu", i), i);
			// identical signals for symbol nodes
			this->snFU[i]->BindCommonPorts(this);
			this->snFU[i]->in_readEnable(this->oldSD_readEnable);
			this->snFU[i]->in_oddIteration(this->oddIteration);
			this->snFU[i]->in_writeEnable(this->newSD_readEnable);
			
			// different signals for symbol nodes
			this->snFU[i]->out_output(this->sn_soft_decision[i]);
			this->snFU[i]->in_input(this->sn_extrinsic_message[i]);
			this->snFU[i]->in_initialization(this->sn_init[i]);
			this->snFU[i]->in_first_extrinsic(this->first_extrinsic[i]);
			this->snFU[i]->in_last_extrinsic(this->last_extrinsic[i]);
			this->snFU[i]->in_save_values(this->sn_output[i]);
			this->snFU[i]->in_oldSD_address(this->oldSD_address[i]);
			this->snFU[i]->in_newSD_address(this->newSD_address[i]);

			this->snFU[i]->in_SD_read(this->soft_decision_permutations[i * 4 + 1]);
			this->snFU[i]->out_SD_read(this->soft_decision_permutations[i * 4]);
			this->snFU[i]->in_SD_write(this->soft_decision_permutations[i * 4 + 3]);
			this->snFU[i]->out_SD_write(this->soft_decision_permutations[i * 4 + 2]);

			this->mux1[i] = new Multiplexer<TYPE >(PosIntToString("muxSnInput", i), this->received_value, this->cn_extrinsic_message_permuted[i], this->sn_init[i], this->sn_extrinsic_message[i]);
			this->mux2[i] = new Multiplexer<TYPE >(PosIntToString("muxSnOutput", i), this->soft_decision_permutations[i * 4 + 1], this->sn_soft_decision_delayed[i], this->oddIteration, this->sn_output_permuted[i]);

			this->delay_of_SN_output[i] = new ShiftRegister<TYPE >(PosIntToString("DelayOfSNoutput_", i), DURATION_PERMUTATION, WIDTH_SD);
			this->delay_of_SN_output[i]->input(this->sn_soft_decision[i]);
			this->delay_of_SN_output[i]->output(this->sn_soft_decision_delayed[i]);
			this->delay_of_SN_output[i]->BindCommonPorts(this);

			this->delay_write_to_output[i] = new ShiftRegister<sc_logic >(PosIntToString("DelayWriteToOutput_", i), 1 + DURATION_SN_OUTPUT, WIDTH_BOOL);
			this->delay_write_to_output[i]->input(this->sn_output[i]);
			this->delay_write_to_output[i]->output(this->write_to_output_enable[i]);
			this->delay_write_to_output[i]->BindCommonPorts(this);
		}

		this->sd_input = new ShiftRegister<TYPE >("InputDelay", 2 + DURATION_ROM_ACCESS, WIDTH_SD);
		this->sd_input->BindCommonPorts(this);
		this->sd_input->input(this->in_received_value);
		this->sd_input->output(this->received_value);

		/* ****************************** Check Node Functional Unit ***************************** */
		this->cnFU = (CheckNodeFunctionalUnit**) malloc(sizeof(CheckNodeFunctionalUnit*) * PARALLELISM);
		for(int i = 0; i < PARALLELISM; i++)
		{
			this->cnFU[i] = new CheckNodeFunctionalUnit(PosIntToString("CnFu", i));

			// identical signals for check nodes
			this->cnFU[i]->BindCommonPorts(this);
			this->cnFU[i]->in_indexIn(this->cn_index_in);
			this->cnFU[i]->in_indexOut(this->cn_index_out);
			this->cnFU[i]->in_lastValue(this->cn_last_value);
			this->cnFU[i]->in_firstIteration(this->cn_first_iteration);
			this->cnFU[i]->in_firstValue(this->cn_first_value);
			this->cnFU[i]->in_variableDelay(this->cn_variableDelay);
			// memory access
			this->cnFU[i]->in_readMEM(this->cn_mem_read);
			this->cnFU[i]->in_addressMEM(this->cn_mem_address);
			this->cnFU[i]->in_enableMEM(this->cn_MEM_enable);

			// different signals for check nodes
			this->cnFU[i]->in_input(this->cn_soft_decision[i]);
			this->cnFU[i]->output(this->cn_extrinsic_message[i]);
			this->cnFU[i]->out_valid(this->cn_valid[i]);
		}

		/* ****************************** Check Node Control Signals ***************************** */
		this->controlCN = new ControlCN("ControlCN");
		this->controlCN->BindCommonPorts(this);
		this->controlCN->in_firstValue(this->ds_cn_first_value_out);
		this->controlCN->in_oddIteration(this->oddIteration);
		this->controlCN->in_variableDelay(this->sn_variableDelay);
		this->controlCN->in_degreeIndex(this->ds_cn_index);
		this->controlCN->in_valid(this->cns_valid);

		this->controlCN->out_firstValue(this->cn_first_value);
		this->controlCN->out_lastValue(this->cn_last_value);
		this->controlCN->out_degreeIndexIn(this->cn_index_in);
		this->controlCN->out_degreeIndexOut(this->cn_index_out);
		this->controlCN->out_variableDelay(this->cn_variableDelay);
		this->controlCN->out_memoryEnable(this->cn_MEM_enable);
		this->controlCN->out_memoryRead(this->cn_mem_read);
		this->controlCN->out_memoryAddress(this->cn_mem_address);
		this->controlCN->out_decodingValid(this->decoding_valid);

		/* ****************************** Permutation Networks ***************************** */
		this->permutation_networks = (PermutationNetwork**) malloc(sizeof(PermutationNetwork*) * 4);
		for(int i = 0; i < 4; i++)
		{
			this->permutation_networks[i] = new PermutationNetwork(PosIntToString("Pi", i));
			this->permutation_networks[i]->BindCommonPorts(this);
			this->permutation_networks[i]->in_index(this->permutation_networks_indices[i]);
			this->permutation_networks[i]->input(this->pi_input[i]);
			this->permutation_networks[i]->output(this->pi_output[i]);
		}
	}
	void PermutationOutputMapping()
	{
		sc_lv<4 * ROM_PERMUTATION_WIDTH> permutationIndices_vector = permutation_networks_indices_vector;
		for(int i = 0; i < 4; i++)
		{
			int max = (4 - i) * ROM_PERMUTATION_WIDTH - 1;
			int min = (4 - i - 1) * ROM_PERMUTATION_WIDTH;
			this->permutation_networks_indices[i].write(permutationIndices_vector(max, min));
		}
	}


	void PermutationOutput()
	{
		for(int j = 0; j < PARALLELISM; j++)
		{
			int max = (PARALLELISM - j) * WIDTH_SD - 1;
			int min = (PARALLELISM - j - 1) * WIDTH_SD;;
			cn_soft_decision[j].write(pi_output[0].read()(max,min));
			cn_extrinsic_message_permuted[j].write(pi_output[1].read()(max,min));
			soft_decision_permutations[j * 4 + 1].write(pi_output[2].read()(max,min));
			soft_decision_permutations[j * 4 + 3].write(pi_output[3].read()(max,min));
		}
	}

	void PermutationInput()
	{
		sc_lv<PARALLELISM * WIDTH_SD> pi_input[4];
		for(int j = 0; j < PARALLELISM; j++)
		{
			int max = (PARALLELISM - j) * WIDTH_SD - 1;
			int min = (PARALLELISM - j - 1) * WIDTH_SD;
			pi_input[0](max, min) = sn_output_permuted[j].read();
			pi_input[1](max, min) = cn_extrinsic_message[j].read();
			pi_input[2](max, min) = soft_decision_permutations[j * 4].read();
			pi_input[3](max, min) = soft_decision_permutations[j * 4 + 2].read();
		}
		this->pi_input[0].write(pi_input[0]);
		this->pi_input[1].write(pi_input[1]);
		this->pi_input[2].write(pi_input[2]);
		this->pi_input[3].write(pi_input[3]);
	}

	void Validity()
	{
		sc_lv<PARALLELISM> temp;
		for(int i = 0; i < PARALLELISM; i++)
		{
			temp[i] = this->cn_valid[i].read();
		}
		this->cns_valid.write(temp);
	}
	
	void RomSNOutput()
	{

		for(int i = 0; i < PARALLELISM; i++)
		{
			this->newSD_address[i].write(this->romSN_output.read().range((PARALLELISM - i)*ROM_SN_WIDTH - 1,(PARALLELISM - i - 1)*ROM_SN_WIDTH + 2));
			this->first_extrinsic[i].write(this->romSN_output.read()[(PARALLELISM - i - 1)*ROM_SN_WIDTH + 1]);
			this->last_extrinsic[i].write(this->romSN_output.read()[(PARALLELISM - i - 1)*ROM_SN_WIDTH]);
		}
	}

	void SNAddress()
	{
		sc_lv<(ROM_SN_WIDTH - 2) * PARALLELISM> oldSD_address_vector = this->oldSD_address_vector.read();
		for(int i = 0; i < PARALLELISM; i++)
		{
			if(this->sn_init[i].read() == SC_LOGIC_1 || this->sn_output[i].read() == SC_LOGIC_1)
			{
				this->oldSD_address[i].write(this->sn_address.read());
			}
			else
			{
				int max = (PARALLELISM - i) * (ROM_SN_WIDTH - 2) - 1;
				int min = (PARALLELISM - i - 1) * (ROM_SN_WIDTH - 2);
				this->oldSD_address[i].write(oldSD_address_vector(max,min));
			}
		}
	}

	void SNInitOutputControl()
	{
		for(int i = 0; i < PARALLELISM; i++)
		{
			this->sn_init[i].write(this->sn_init_vector.read()[i]);
			this->sn_output[i].write(this->sn_output_vector.read()[i]);
		}
	}

	void Update()
	{
		this->out_decoded_value_present.write(SC_LOGIC_0);		
		// Determination of HDs and Writing output (1-Bit Sign-Register with enable)
		for(int i = 0; i < PARALLELISM; i++)
		{
			if(this->write_to_output_enable[i].read() == SC_LOGIC_1)
			{
				this->out_decoded_value_present.write(SC_LOGIC_1);
				this->out_decoded_value.write(this->sn_soft_decision[i].read()[WIDTH_SD-1]);
			}
		}

		// Setting variable delay
		if(this->ds_cn_first_value_out.read() == true)
		{
			this->sn_variableDelay.write(this->ds_cn_degree.read());
		}
	}

	void Reset()
	{
		this->sn_variableDelay.write(1);
	}
};
#else
class ParallelOPMP: public VHDL_Component
{
public:
	sc_in<TYPE > in_received_value;						// received SD-Values
	sc_in<sc_logic> in_new_values_present;				// this flag determines wether the next received code word is available
	sc_out<sc_logic> out_ready_for_next_received_value;	// this flag indicates, that the decoder ist ready for the next received symbol
	sc_out<sc_lv<WIDTH_ITERATION> > out_iterations;
	sc_out<sc_logic> out_decoded_value;					// decoded HD-Values
	sc_out<sc_logic> out_decoded_value_present;
	sc_out<sc_logic> out_decoding_valid;
	ParallelOPMP(sc_module_name nm, const char* degreeFileName, int max_iterations) : VHDL_Component(nm),
		in_received_value("in_received_value"),
		in_new_values_present("in_new_values_present"),
		out_ready_for_next_received_value("out_ready_for_next_received_value"),
		out_iterations("out_iterations"),
		out_decoded_value("out_decoded_value"),
		out_decoded_value_present("out_decoded_value_present"),
		out_decoding_valid("out_decoding_valid")
    {
		elaborate_foreign_module("work.Decoder");
    }
};
#endif // MTI_SYSTEMC

#endif //PARALLELOPMP_H
