#ifndef MULTIPLEXER_H
#define MULTIPLEXER_H

template<class T>class Multiplexer: public sc_module
{
public:
	sc_in<T> inputA;
	sc_in<T> inputB;
	sc_in<sc_logic> select;
	sc_out<T> output;
	
	SC_HAS_PROCESS(Multiplexer);
	Multiplexer(sc_module_name name) :
		inputA("inputA"), inputB("inputB"), select("select"), output("output")
	{
		SC_METHOD(Update);
		sensitive << this->inputA << this->inputB << this->select;
	}
	Multiplexer(sc_module_name name, sc_signal<T> &inputA, sc_signal<T> &inputB, sc_signal<sc_logic> &select, sc_signal<T> &output) :
		inputA("inputA"), inputB("inputB"), select("select"), output("output")
	{
		SC_METHOD(Update);
		sensitive << this->inputA << this->inputB << this->select;
		this->inputA.bind(inputA);
		this->inputB.bind(inputB);
		this->select.bind(select);
		this->output.bind(output);
	}

	void Update()
	{
		output.write(0);
		if(select.read() == SC_LOGIC_1)
		{
			output.write(inputA.read());
		}
		else if(select.read() == SC_LOGIC_0)
		{
			output.write(inputB.read());
		}
	}
};
#endif // MULTIPLEXER_H
