#ifndef PERMUTATIONNETWORK_H
#define PERMUTATIONNETWORK_H
#include "..\Component.h"

#ifndef MTI_SYSTEMC
class PermutationNetwork : public Component
{
public:
	sc_in<sc_lv<PARALLELISM * WIDTH_SD> > input;
	sc_out<sc_lv<PARALLELISM * WIDTH_SD> > output;
	sc_in<sc_lv<ROM_PERMUTATION_WIDTH> > in_index;

	PermutationNetwork(sc_module_name name)
	{
	}

	void Reset()
	{
		for(int i = 0; i < PARALLELISM; i++)
		{
			this->output.write(0);
		}
	}

	void Update()
	{
		const char* name = this->name();
		sc_lv<ROM_PERMUTATION_WIDTH> index = in_index.read();
		sc_lv<PARALLELISM * WIDTH_SD> input = this->input;
		sc_lv<PARALLELISM * WIDTH_SD> output = this->output;
		for(int i = 0; i < PARALLELISM; i++)
		{
			int max1 = (PARALLELISM - i) * WIDTH_SD - 1;
			int min1 = (PARALLELISM - i - 1) * WIDTH_SD;
			int temp = index.range((PARALLELISM - i) * PARALLELISM_WIDTH - 1, (PARALLELISM - i - 1) * PARALLELISM_WIDTH).to_uint();
			int max2 = (PARALLELISM - temp) * WIDTH_SD - 1;
			int min2 = (PARALLELISM - temp - 1) * WIDTH_SD;
			output(max1, min1) = input(max2, min2);			
			this->output.write(output);
		}
	}
};
#else
class PermutationNetwork : public VHDL_Component
{
public:
	sc_in<sc_lv<PARALLELISM * WIDTH_SD> > input;
	sc_out<sc_lv<PARALLELISM * WIDTH_SD> > output;
	sc_in<sc_lv<ROM_PERMUTATION_WIDTH> > in_index;

	PermutationNetwork(sc_module_name name) : VHDL_Component(name),
		input("in_input"),
		output("out_output"),
		in_index("in_index")
	{
		elaborate_foreign_module("work.PermutationNetwork");
	}
};
#endif //MTI_SYSTEMC

#endif //PERMUTATIONNETWORK_H
