#ifndef CONTROLCN_H
#define CONTROLCN_H
#include "..\Component.h"
#include "..\..\..\ControlInformation\widths.h"

#ifndef MTI_SYSTEMC
class ControlCN : public Component
{
public:
	sc_in<sc_logic>						in_firstValue;
	sc_in<sc_logic>						in_oddIteration;
	sc_in<sc_lv<DEGREE_WIDTH> >			in_variableDelay;
	sc_in<sc_lv<DEGREE_WIDTH> >			in_degreeIndex;
	sc_in<sc_lv<PARALLELISM> >			in_valid;

	sc_out<sc_logic>					out_firstValue;
	sc_out<sc_logic>					out_lastValue;
	sc_out<sc_lv<DEGREE_WIDTH> >		out_degreeIndexIn;
	sc_out<sc_lv<DEGREE_WIDTH> >		out_degreeIndexOut;
	sc_out<sc_lv<DEGREE_WIDTH> >		out_variableDelay;
	sc_out<sc_logic>					out_memoryEnable;
	sc_out<sc_logic>					out_memoryRead;
	sc_out<sc_lv<RAM_CN_ADDRESS_WIDTH> >out_memoryAddress;
	sc_out<sc_logic>					out_decodingValid;

	// Memory Arbitration
	sc_signal<sc_logic> simultanMemoryAccess;
	sc_signal<sc_logic> lastValue_delayed;
	sc_signal<sc_logic> firstValue;

	// Check Node Memory Access
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > readAddress;
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > readAddressCounter;
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > readAddressCounter_next;
	
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > writeAddress;
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > writeAddressCounter;
	sc_signal<sc_lv<RAM_CN_ADDRESS_WIDTH> > writeAddressCounter_next;

	// Validation
	sc_signal<sc_logic> decodingValid;
	sc_signal<int> errors;

	// Delays
	ShiftRegister<sc_logic>* delay_of_CN_first_value;
	ShiftRegister<sc_logic>* shift1;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >* shift2;
	ShiftRegister<sc_logic>* shift4;
	VariableShiftRegister<sc_logic>* variableShift1;
	VariableShiftRegister<sc_lv<DEGREE_WIDTH> >* variableShift2;
	ShiftRegister<sc_lv<DEGREE_WIDTH> >* delayShift;
	ShiftRegister<sc_logic>* shift3;

	SC_HAS_PROCESS(ControlCN);
	ControlCN(sc_module_name name)
	{
		SC_METHOD(RamCNAddress);
		sensitive << this->out_memoryEnable << this->out_memoryRead << this->readAddress << this->writeAddress;
		SC_METHOD(NextAddresses);
		sensitive << this->readAddressCounter << this->writeAddressCounter << this->in_oddIteration;

		// See CN_MEM_Arb for the check node delay (delay between firstValue and firstValue_delayed)
		this->delay_of_CN_first_value = new ShiftRegister<sc_logic>("CN_firstValue_delay", 1, WIDTH_BOOL); // fixed value
		this->delay_of_CN_first_value->input(this->firstValue);
		this->delay_of_CN_first_value->output(this->out_memoryRead);
		this->delay_of_CN_first_value->BindCommonPorts(this);
		this->shift1 = new ShiftRegister<sc_logic>("Shift1", 3, WIDTH_BOOL);// fixed value
		this->shift1->input(this->out_memoryRead);
		this->shift1->output(this->out_firstValue); 
		this->shift1->BindCommonPorts(this);
		this->shift4 = new ShiftRegister<sc_logic>("Shift4", DELAY_CN - 1, WIDTH_BOOL);
		this->shift4->input(this->in_firstValue);
		this->shift4->output(this->firstValue);
		this->shift4->BindCommonPorts(this);
		this->delayShift = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("DelayShift", DELAY_CN - 1, DEGREE_WIDTH);
		this->delayShift->input(this->in_variableDelay);
		this->delayShift->output(this->out_variableDelay);
		this->delayShift->BindCommonPorts(this);
		this->shift2 = new ShiftRegister<sc_lv<DEGREE_WIDTH> >("Shift2", DELAY_CN, DEGREE_WIDTH);
		this->shift2->input(this->in_degreeIndex);
		this->shift2->output(this->out_degreeIndexIn);
		this->shift2->BindCommonPorts(this);

		this->variableShift1 = new VariableShiftRegister<sc_logic>("VariableShift1", MAX_DEGREE, 0, 1);
		this->variableShift1->BindCommonPorts(this);
		this->variableShift1->input(this->out_firstValue);
		this->variableShift1->output(this->out_lastValue);
		this->variableShift1->active_size(this->out_variableDelay);
		this->variableShift2 = new VariableShiftRegister<sc_lv<DEGREE_WIDTH> >("VariableShift2", MAX_DEGREE, DURATION_CN, DEGREE_WIDTH);
		this->variableShift2->BindCommonPorts(this);
		this->variableShift2->input(this->out_degreeIndexIn);
		this->variableShift2->output(this->out_degreeIndexOut);
		this->variableShift2->active_size(this->out_variableDelay);

		// Validation Delay
		this->shift3 = new ShiftRegister<sc_logic>("Shift3", MAX_DEGREE + 2 * DURATION_PERMUTATION + 2, WIDTH_BOOL);
		this->shift3->input(this->decodingValid);
		this->shift3->output(this->out_decodingValid);
		this->shift3->BindCommonPorts(this);
	}

	void RamCNAddress()
	{
		if(this->out_memoryEnable.read() == SC_LOGIC_1 && this->out_memoryRead.read() == SC_LOGIC_0)
		{
			this->out_memoryAddress.write(this->writeAddress.read());
		}
		else
		{
			this->out_memoryAddress.write(this->readAddress.read());
		}
	}

	void NextAddresses()
	{
		if(in_oddIteration.read() == SC_LOGIC_0)
		{
			this->readAddressCounter_next.write(this->readAddressCounter.read().to_uint() + 1);
			this->writeAddressCounter_next.write(this->writeAddressCounter.read().to_uint() + 1);
		}
		else
		{
			this->readAddressCounter_next.write(this->readAddressCounter.read().to_uint() - 1);
			this->writeAddressCounter_next.write(this->writeAddressCounter.read().to_uint() - 1);
		}
	}

	void Update()
	{
		// Arbitration of CN-MEM-Access
		this->lastValue_delayed.write(this->out_lastValue.read());
		this->simultanMemoryAccess.write(this->firstValue.read() & this->out_lastValue.read());
		this->out_memoryEnable.write(this->simultanMemoryAccess.read() == SC_LOGIC_1 ? SC_LOGIC_1 : this->out_lastValue.read() | this->firstValue.read());

		// Generation of read and write addresses
		if(this->firstValue.read() == SC_LOGIC_1)
		{
			this->readAddressCounter.write(this->readAddressCounter_next.read());
			if(this->in_oddIteration.read() == SC_LOGIC_0)
			{
				this->readAddress.write(this->readAddressCounter_next.read());
			}
			else if(this->in_oddIteration.read() == SC_LOGIC_1)
			{
				this->readAddress.write(this->readAddressCounter.read());
			}
		}
		if(this->out_lastValue.read() == SC_LOGIC_1)
		{
			this->writeAddressCounter.write(writeAddressCounter_next.read());
			if(this->in_oddIteration.read() == SC_LOGIC_0)
			{
				this->writeAddress.write(writeAddressCounter_next.read());
			}
			else if(this->in_oddIteration.read() == SC_LOGIC_1)
			{
				this->writeAddress.write(this->writeAddressCounter.read());
			}
		}

		// Validation
		this->decodingValid.write(SC_LOGIC_0);
		if(this->in_oddIteration.read() == SC_LOGIC_1)
		{
			if(this->lastValue_delayed.read() == SC_LOGIC_1)
			{
				int errors = this->errors.read();
				sc_lv<PARALLELISM> temp = this->in_valid.read();
				for(int i = 0; i < PARALLELISM; i++)
				{
					if(temp[i] == SC_LOGIC_0)
					{
						errors++;
					}
				}
				this->errors.write(errors);
				if(this->out_memoryAddress.read().to_uint() == 0)
				{
					if(errors == 0)
					{
						this->decodingValid.write(SC_LOGIC_1);
					}
					this->errors.write(0);
				}
			}
		}
	}

	void Reset()
	{
		sc_lv<RAM_CN_ADDRESS_WIDTH> max;
		for(int i = 0; i < RAM_CN_ADDRESS_WIDTH; i++)
		{
			max[i] = SC_LOGIC_1;
		}
		this->readAddress.write(max);
		this->writeAddress.write(max);
		this->readAddressCounter.write(max);
		this->writeAddressCounter.write(max);
	}
};
#else
class ControlCN : public VHDL_Component
{
public:
	sc_in<sc_logic>						in_firstValue;
	sc_in<sc_logic>						in_oddIteration;
	sc_in<sc_lv<DEGREE_WIDTH> >			in_variableDelay;
	sc_in<sc_lv<DEGREE_WIDTH> >			in_degreeIndex;
	sc_in<sc_lv<PARALLELISM> >			in_valid;

	sc_out<sc_logic>					out_firstValue;
	sc_out<sc_logic>					out_lastValue;
	sc_out<sc_lv<DEGREE_WIDTH> >		out_degreeIndexIn;
	sc_out<sc_lv<DEGREE_WIDTH> >		out_degreeIndexOut;
	sc_out<sc_lv<DEGREE_WIDTH> >		out_variableDelay;
	sc_out<sc_logic>					out_memoryEnable;
	sc_out<sc_logic>					out_memoryRead;
	sc_out<sc_lv<RAM_CN_ADDRESS_WIDTH> >out_memoryAddress;
	sc_out<sc_logic>					out_decodingValid;
	    
	ControlCN(sc_module_name nm) : VHDL_Component(nm),
		in_firstValue("in_firstValue"),
		in_oddIteration("in_oddIteration"),
		in_variableDelay("in_variableDelay"),
		in_degreeIndex("in_degreeIndex"),
		in_valid("in_valid"),

		out_firstValue("out_firstValue"),
		out_lastValue("out_lastValue"),
		out_degreeIndexIn("out_degreeIndexIn"),
		out_degreeIndexOut("out_degreeIndexOut"),
		out_variableDelay("out_variableDelay"),
		out_memoryEnable("out_memoryEnable"),
		out_memoryRead("out_memoryRead"),
		out_memoryAddress("out_memoryAddress"),
		out_decodingValid("out_valid")
    {
		elaborate_foreign_module("work.ControlCN");
    }
};
#endif // MTI_SYSTEMC

#endif //CONTROLCN_H