#ifndef INITWRITEBACK_H
#define INITWRITEBACK_H
#include "..\Component.h"
#include "..\Components\Memory.h"
#include "..\Components\VariableShiftRegister.h"

#ifndef MTI_SYSTEMC
class InitAndOutput : public Component
{
public:
	sc_in<sc_logic>					in_init;
	sc_in<sc_logic>					in_writeBack;
	sc_out<sc_logic>				out_loadAndwriteBackDone;
	sc_out<sc_lv<PARALLELISM> >		out_writeBack;
	sc_out<sc_lv<PARALLELISM> >		out_init;
	sc_out<sc_lv<ROM_SN_WIDTH-2> >	out_address;
	
	RomMemory2<SN_NUMBER, (ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH, ROM_INITILIZATION_ADDRESS_WIDTH>*	controlMemory;
	sc_signal<sc_lv<ROM_INITILIZATION_ADDRESS_WIDTH> >												controlMEM_address;
	sc_signal<sc_logic>																				controlMEM_enable;
	sc_signal<sc_lv<(ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH> >										controlMEM_output;

	sc_signal<sc_logic>				init;
	sc_signal<sc_logic>				writeBack;
	sc_signal<int>					counter;
	ShiftRegister<sc_logic >* delay_init;
	ShiftRegister<sc_logic >* delay_writeBack;

	SC_HAS_PROCESS(InitAndOutput);
	InitAndOutput(sc_module_name name)
	{
		SC_METHOD(MemEnable);
		sensitive << this->in_init << this->in_writeBack;

		SC_METHOD(MemAddress);
		sensitive << counter;

		this->controlMemory = new RomMemory2<SN_NUMBER, (ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH, ROM_INITILIZATION_ADDRESS_WIDTH>("RomInit");
		this->controlMemory->BindCommonPorts(this);
		this->controlMemory->readEnable(this->controlMEM_enable);
		this->controlMemory->readAddress(this->controlMEM_address);
		this->controlMemory->output(this->controlMEM_output);
		this->controlMEM_output.write(0);

		this->delay_init = new ShiftRegister<sc_logic>("Delay_Init", DURATION_ROM_ACCESS, 1);
		this->delay_init->BindCommonPorts(this);
		this->delay_init->input(this->in_init);
		this->delay_init->output(this->init);

		this->delay_writeBack = new ShiftRegister<sc_logic>("Delay_WriteBack", DURATION_ROM_ACCESS, 1);
		this->delay_writeBack->BindCommonPorts(this);
		this->delay_writeBack->input(this->in_writeBack);
		this->delay_writeBack->output(this->writeBack);
	}

	void MemEnable()
	{
		this->controlMEM_enable.write(this->in_init.read() | this->in_writeBack.read());
	}

	void MemAddress()
	{
		int index = this->counter.read();
		this->controlMEM_address.write(index);
	}

	void Update()
	{
		// preload
		this->out_loadAndwriteBackDone = SC_LOGIC_0;
		this->out_init.write(0);
		this->out_writeBack.write(0);

		// Output of ROM
		sc_lv<(ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH> temp = this->controlMEM_output.read();
		this->out_address.write(temp.range(ROM_SN_WIDTH - 3, 0));
		int bank = temp.range((ROM_SN_WIDTH - 2) + PARALLELISM_WIDTH - 1, ROM_SN_WIDTH - 2).to_uint();
		sc_lv<PARALLELISM> banks = 0;
		banks[bank] = SC_LOGIC_1;
		if(this->init.read() == SC_LOGIC_1)
		{
			this->out_init.write(banks);
		}
		if(this->writeBack.read() == SC_LOGIC_1)
		{
			this->out_writeBack.write(banks);
		}

		// counter
		if((this->in_init.read() | this->in_writeBack.read()) == SC_LOGIC_1)
		{
			int index = this->counter.read();
			if(index == SN_NUMBER - 3)
			{
				this->out_loadAndwriteBackDone = SC_LOGIC_1;
			}

			
			if(index == SN_NUMBER - 1)
			{
				this->counter.write(0);
			}
			else
			{
				index++;
				this->counter.write(index);
			}
		}
	}

	void Reset()
	{
		this->counter.write(0);
	}
};
#else
class InitAndOutput : public VHDL_Component
{
public:
	sc_in<sc_logic>					in_init;
	sc_in<sc_logic>					in_writeBack;
	sc_out<sc_logic>				out_loadAndwriteBackDone;
	sc_out<sc_lv<PARALLELISM> >		out_writeBack;
	sc_out<sc_lv<PARALLELISM> >		out_init;
	sc_out<sc_lv<ROM_SN_WIDTH-2> >	out_address;

	InitAndOutput(sc_module_name name) : VHDL_Component(name),
		in_init("in_initState"),
		in_writeBack("in_outputState"),
		out_loadAndwriteBackDone("out_loadAndOutputDone"),
		out_writeBack("out_output"),
		out_address("out_address"),
		out_init("out_init")		
	{
		elaborate_foreign_module("work.InitAndOutput");
	}
};
#endif // MTI_SYSTEMC

#endif // INITWRITEBACK_H