#ifndef DECODERMONITOR_H
#define DECODERMONITOR_H
#include "..\Decoder\Architectures\ParallelOPMP.h"
#include "Monitor.h"
#define WIDTH_RECEIVED WIDTH_SD - 1

class DecoderMonitor: public Monitor
{
public:

	// received values
	sc_signal<TYPE > received_values;
	sc_signal<sc_logic> new_values_present;					// indicates to the decoder that new values are present
	sc_signal<sc_logic> decoder_ready_for_next_codeword;	// decoder ready for new values to decode
	// decoded values
	sc_signal<sc_logic> decoded_values;
	sc_signal<sc_logic> decoded_value_present;				// is true after the decoding despite the success
	sc_signal<sc_logic> decoding_valid;						// indicates success of the decoding
	sc_signal<sc_lv<WIDTH_ITERATION> > iterations;

	float sdValue;
	sc_fixed<WIDTH_RECEIVED, WIDTH_RECEIVED - PRECISION_WIDTH, SC_TRN, SC_SAT> fixedPoint;
	sc_lv<WIDTH_RECEIVED> logicVector;
	int integerRepresentation;
	
	int length;
	bool* codeword;
	int errors_before;
	int errors_before_last;
	int errors_before_last2;
	int position_SD;
	int position_HD;
	int errors_after;
	bool terminate_simulation;
	int maxReceived;
	int minReceived;

	ParallelOPMP* decoder;
	FILE* file_received;
	DecoderMonitor(sc_module_name name) : Monitor(name)
	{
		int max_iter = 100;
		Calculate_MinMaxValues();

		// open file with received symbols
		//if (fopen_s(&this->file_received, "..\\ControlInformation\\SerialPCMregLDPC.received","r") != NULL)
#ifdef MTI_SYSTEMC
		this->decoder = new ParallelOPMP("OPMP", "..\\ControlInformation\\RomDS.mif", max_iter);
		this->file_received = fopen( "..\\ControlInformation\\1.received","r");
#else
		if (fopen_s(&this->file_received, "..\\ControlInformation\\1.received","r") != NULL) 
		{ 
			printf("Failed to open file containing the received symbols.\n"); 
		}
		this->decoder = new ParallelOPMP("OPMP", "..\\ControlInformation\\RomDS.mif", max_iter);
#endif //MTI_SYSTEMC

		fseek(this->file_received, 0L, SEEK_SET);
		fscanf(this->file_received,"%i", &this->length);
		this->codeword = new bool[this->length];
		for(int i = 0; i < this->length; i++)
		{
			int temp;
			fscanf(this->file_received,"%i", &temp);
			if(temp < 0)
			{
				this->codeword[i] = true;
			}
			else
			{
				this->codeword[i] = false;
			}
		}

		this->decoder->clock.bind(this->iclock);
		this->decoder->enable.bind(this->enable);
		this->decoder->reset.bind(this->reset);

		this->decoder->in_received_value.bind(this->received_values);
		this->decoder->out_decoded_value.bind(this->decoded_values);
		this->decoder->out_decoded_value_present.bind(this->decoded_value_present);
		this->decoder->out_decoding_valid.bind(this->decoding_valid);

		this->decoder->in_new_values_present.bind(this->new_values_present);
		this->decoder->out_ready_for_next_received_value.bind(this->decoder_ready_for_next_codeword);
		this->decoder->out_iterations(this->iterations);
		this->position_SD = 0;
		this->errors_before = 0;
		this->position_HD = 0;
		this->errors_after = 0;
		this->terminate_simulation = false;

		sc_lv<WIDTH_RECEIVED - 1> lv;
		for(int i = 0; i < WIDTH_RECEIVED - 1; i++)
		{
			lv[i] = SC_LOGIC_1;
		}
		maxReceived = lv.to_uint();
		minReceived = -maxReceived;
		new_values_present = SC_LOGIC_0;
	}

	void TraceSignals(sc_trace_file *tf)
	{
#ifndef MTI_SYSTEMC
		sc_trace(tf, this->decoder->decodeScheduling->out_rom_index, "out_rom_index");
		sc_trace(tf, this->decoder->decodeScheduling->out_rom_enable, "out_mem_enable");
		sc_trace(tf, this->decoder->decodeScheduling->out_first_value, "out_first_value");
		sc_trace(tf, this->decoder->decodeScheduling->out_cn_index, "out_cn_index");
		sc_trace(tf, this->decoder->decodeScheduling->out_cn_degree, "out_cn_degree");
		sc_trace(tf, this->decoder->decodeScheduling->out_iteration, "out_iteration");
		sc_trace(tf, this->decoder->decodeScheduling->enable, "in_enable");
		sc_trace(tf, this->decoder->decodeScheduling->reset, "in_reset");

		sc_trace(tf, this->decoder->controlSNandPI->in_oddIteration, "SNandPI->in_oddIteration");
		sc_trace(tf, this->decoder->controlSNandPI->in_enableRom, "SNandPI->in_enableRom");
		sc_trace(tf, this->decoder->controlSNandPI->in_addressRom, "SNandPI->in_addressRom");
		sc_trace(tf, this->decoder->controlSNandPI->in_variableDelay, "SNandPI->in_variableDelay");
		sc_trace(tf, this->decoder->controlSNandPI->out_newSD_readEnable, "SNandPI->out_newSD_readEnable");
		sc_trace(tf, this->decoder->controlSNandPI->out_oldSD_readEnable, "SNandPI->out_oldSD_readEnable");
		sc_trace(tf, this->decoder->controlSNandPI->out_oldSD_address_vector, "SNandPI->out_oldSD_address_vector");
		sc_trace(tf, this->decoder->controlSNandPI->out_newSD_address_vector, "SNandPI->out_newSD_address_vector");
		sc_trace(tf, this->decoder->controlSNandPI->out_permutationIndices_vector, "SNandPI->out_permutationIndices_vector");
		sc_trace(tf, this->decoder->controlSNandPI->romSN_output, "SNandPI->romSN_output");
		sc_trace(tf, this->decoder->controlSNandPI->banks_permutation_MEM->output, "SNandPI->Pi2output");
		sc_trace(tf, this->decoder->controlSNandPI->nodes_permutation_MEM->output, "SNandPI->Pi0output");
		sc_trace(tf, this->decoder->controlSNandPI->permutationIndices[0], "permutationIndices[0]");
		sc_trace(tf, this->decoder->controlSNandPI->permutationIndices[1], "permutationIndices[1]");
		sc_trace(tf, this->decoder->controlSNandPI->permutationIndices[2], "permutationIndices[2]");
		sc_trace(tf, this->decoder->controlSNandPI->permutationIndices[3], "permutationIndices[3]");


		/*
		sc_trace(tf, this->decoder->cn_mem_address, "cn_mem_address");
		sc_trace(tf, this->decoder->cn_MEM_enable, "cn_MEM_enable");
		sc_trace(tf, this->decoder->ds_cn_first_value_delayed, "ds_cn_first_value_delayed");
		sc_trace(tf, this->decoder->cn_mem_address2, "cn_mem_address2");
		sc_trace(tf, this->decoder->cn_readAddress, "cn_readAddress");
		sc_trace(tf, this->decoder->cn_writeAddress, "cn_writeAddress");
		sc_trace(tf, this->decoder->cn_readAddressCounter, "cn_readAddressCounter");
		sc_trace(tf, this->decoder->cn_writeAddressCounter, "cn_writeAddressCounter");
		*/
		/*
		// Decode Scheduling Outputs
		sc_trace(tf, this->decoder->ds_control_MEM_address, "ds_control_MEM_address");
		sc_trace(tf, this->decoder->ds_MEM_enable, "ds_MEM_enable");
		sc_trace(tf, this->decoder->ds_cn_degree, "ds_cn_degree");
		sc_trace(tf, this->decoder->ds_cn_index, "ds_cn_index");
		sc_trace(tf, this->decoder->ds_cn_first_value, "ds_cn_first_value");
		sc_trace(tf, this->decoder->iteration, "iteration");
		sc_trace(tf, this->decoder->ds_cn_mem_address, "ds_cn_mem_address");

		// Check Node
		//sc_trace(tf, this->decoder->cnFU[0]->extrinsic_memory_output, "extrinsic_memory_output");
		sc_trace(tf, this->decoder->cnFU[0]->in_indexIn, "cnFU[0]->in_indexIn");
		sc_trace(tf, this->decoder->cnFU[0]->in_indexOut, "cnFU[0]->in_indexOut");
		sc_trace(tf, this->decoder->cnFU[0]->in_firstValue, "cnFU[0]->in_firstValue");
		sc_trace(tf, this->decoder->cnFU[0]->in_lastValue, "cnFU[0]->in_lastValue");
		sc_trace(tf, this->decoder->cnFU[0]->in_firstIteration, "cnFU[0]->in_firstIteration");
		sc_trace(tf, this->decoder->cnFU[0]->in_variableDelay, "cnFU[0]->in_variableDelay");
		sc_trace(tf, this->decoder->cnFU[0]->in_enableMEM, "cnFU[0]->in_enableMEM");
		sc_trace(tf, this->decoder->cnFU[0]->in_readMEM, "cnFU[0]->in_readMEM");
		sc_trace(tf, this->decoder->cnFU[0]->in_addressMEM, "cnFU[0]->in_addressMEM");
		sc_trace(tf, this->decoder->cnFU[0]->out_valid, "cnFU[0]->out_valid");*/
		
		// Symbol Nodes
		/*
		sc_trace(tf, this->decoder->snFU[0]->in_oldSD_address, "snFU[0]->readAddress");
		sc_trace(tf, this->decoder->snFU[0]->in_readEnable, "snFU[0]->in_readEnable");
		sc_trace(tf, this->decoder->snFU[1]->in_oldSD_address, "snFU[1]->readAddress");
		sc_trace(tf, this->decoder->snFU[1]->in_readEnable, "snFU[1]->in_readEnable");
		sc_trace(tf, this->decoder->snFU[2]->in_oldSD_address, "snFU[2]->readAddress");
		sc_trace(tf, this->decoder->snFU[2]->in_readEnable, "snFU[2]->in_readEnable");
		//sc_trace(tf, this->decoder->snFU[2]->in_address, "snFU[2]->in_address");
		sc_trace(tf, this->decoder->snFU[2]->in_oldSD_address, "snFU[2]->in_oldSD_address");
		sc_trace(tf, this->decoder->fsm_sn_init[2], "snFU[2]->fsm_sn_init[2]");
		sc_trace(tf, this->decoder->fsm_write_to_output_enable[2], "snFU[2]->fsm_write_to_output_enable[2]");
		*/
		//sc_trace(tf, this->decoder->sn_writeEnable, "sn_writeEnable");
		sc_trace(tf, this->decoder->cnFU[0]->in_input, "cnFU[0]->in_input");
		sc_trace(tf, this->decoder->cnFU[1]->in_input, "cnFU[1]->in_input");
		sc_trace(tf, this->decoder->cnFU[2]->in_input, "cnFU[2]->in_input");
		sc_trace(tf, this->decoder->cnFU[0]->in_firstValue, "cn_firstValue");
		sc_trace(tf, this->decoder->cnFU[0]->in_lastValue, "cn_lastValue");
		sc_trace(tf, this->decoder->cnFU[0]->output, "cnFU[0]->output");
		//sc_trace(tf, this->decoder->permutation_index, "permutation_index");
		sc_trace(tf, this->decoder->snFU[0]->in_SD_read, "snFU[0]->in_SD_read");
		sc_trace(tf, this->decoder->snFU[0]->in_SD_write, "snFU[0]->in_SD_write");
		sc_trace(tf, this->decoder->snFU[0]->out_SD_read, "snFU[0]->out_SD_read");
		sc_trace(tf, this->decoder->snFU[0]->out_SD_write, "snFU[0]->out_SD_write");
		sc_trace(tf, this->decoder->snFU[0]->out_output, "snFU[0]->output");
		sc_trace(tf, this->decoder->snFU[0]->in_input, "snFU[0]->input");
		sc_trace(tf, this->decoder->snFU[1]->in_input, "snFU[1]->input");
		sc_trace(tf, this->decoder->snFU[2]->in_input, "snFU[2]->input");
		sc_trace(tf, this->decoder->snFU[2]->out_output, "snFU[2]->out_output");
		sc_trace(tf, this->decoder->snFU[1]->out_output, "snFU[1]->out_output");
		sc_trace(tf, this->decoder->snFU[0]->out_output, "snFU[0]->out_output");

		sc_trace(tf, this->decoder->cnFU[0]->out_valid, "cnFU[0]->out_valid");
		sc_trace(tf, this->decoder->cnFU[1]->out_valid, "cnFU[1]->out_valid");
		sc_trace(tf, this->decoder->cnFU[2]->out_valid, "cnFU[2]->out_valid");
		sc_trace(tf, this->decoder->cn_last_value_delayed, "cn_last_value_delayed");
		sc_trace(tf, this->decoder->cn_mem_address, "cn_mem_address");
		sc_trace(tf, this->decoder->cnFU[0]->in_variableDelay, "cnFU[0]->in_variableDelay");
		sc_trace(tf, this->decoder->snFU[0]->in_readEnable, "snFU[0]->in_readEnable");
		sc_trace(tf, this->decoder->snFU[0]->in_writeEnable, "snFU[0]->in_writeEnable");
		sc_trace(tf, this->decoder->sn_variableDelay, "sn_variableDelay");
		sc_trace(tf, this->decoder->oddIteration, "oddIteration");
		#endif // MTI_SYSTEMC
	}

	void CyclicUpdate()
	{
		// Write SDs
		if((this->decoder_ready_for_next_codeword.read() | this->position_SD != 1) == SC_LOGIC_1)
		{
			if(fscanf(this->file_received,"%f", &sdValue) == EOF)
			{
				this->new_values_present.write(SC_LOGIC_0);
			}
			else
			{
				fixedPoint = sdValue;
				logicVector = static_cast<sc_bv_base>(fixedPoint());
				integerRepresentation = logicVector.to_int();
				integerRepresentation = CheckLimits(integerRepresentation);
				if(integerRepresentation > maxReceived)
				{
					integerRepresentation = maxReceived;
				}
				else if(integerRepresentation < minReceived)
				{
					integerRepresentation = minReceived;
				}
			
				this->new_values_present.write(SC_LOGIC_1);
				this->received_values.write(integerRepresentation);
				bool sign = sdValue < 0 ? true : false;
				if(sign != this->codeword[position_SD])
				{
					errors_before++;
				}
				position_SD++;
				if(position_SD == this->length)
				{
					this->position_SD = 0;
					this->errors_before_last2 = this->errors_before_last;
					this->errors_before_last = this->errors_before;
					this->errors_before = 0;
				}
			}
		}

		// Read HDs
		if(this->decoded_value_present.read() == true)
		{
			bool temp = this->decoder->out_decoded_value.read().to_bool();
			if(temp != this->codeword[position_HD])
			{
				errors_after++;
			}
			position_HD++;
			if(position_HD == this->length)
			{
				if(this->terminate_simulation == true)
				{
					cout << "Predecoded errors: " << this->errors_before_last << "\t";
				}
				else
				{
					cout << "Predecoded errors: " << this->errors_before_last2 << "\t";
				}
				if(this->decoding_valid.read() == SC_LOGIC_1)
				{
					cout << "Decoding valid after "<< this->iterations.read().to_uint() << " iterations with " << errors_after << " errors.\n";
				}
				else
				{
					cout << "Decoding failed after "<< this->iterations.read().to_uint() << " iterations with " << errors_after << " errors.\n";
				}
				errors_after = 0;
				position_HD = 0;
				//sc_stop();
				// Termination of simulation
				if(this->terminate_simulation == true)
				{
					#ifdef MTI_SYSTEMC
						cout << "Duration: " << sc_time_stamp().value() / 20 << " cycles." << endl;
					#else
						cout << "Duration: " << sc_time_stamp().value() / 20000 << " cycles." << endl;
					#endif //MTI_SYSTEMC
					sc_stop();
				}
				if(this->new_values_present.read() == false)
				{
					this->terminate_simulation = true;
				}
			}
		}
	}

	void SequentialUpdate()
	{
	}
};

#endif //DECODERMONITOR_H
