#ifndef MONITOR_H
#define MONITOR_H

#include "..\includes.h"

class Monitor : public sc_module
{
public:
	sc_in<bool> clock;
	sc_signal<sc_logic> iclock;
	sc_signal<sc_logic> reset;
	sc_signal<sc_logic> enable;

	SC_HAS_PROCESS(Monitor);
	Monitor(sc_module_name name)
	{
		SC_METHOD(ClockUpdate);
		sensitive << clock;
		SC_CTHREAD(Update, this->iclock);
		SC_CTHREAD(GenStimuli, this->iclock);
		enable = SC_LOGIC_0;
		reset = SC_LOGIC_0;
	}

	void ClockUpdate()
	{
		if(clock)
			iclock.write(SC_LOGIC_1);
		else
			iclock.write(SC_LOGIC_0);
	}

	void GenStimuli()
	{
		wait(2);
		reset = SC_LOGIC_1;
		wait();
		reset = SC_LOGIC_0;
		wait();
		//wait();
		enable = SC_LOGIC_1;
		this->SequentialUpdate();
	}

	void ConnectToVCD(sc_trace_file *tf)
	{
		sc_trace(tf, clock, "clock");
		sc_trace(tf, enable, "enable");
		sc_trace(tf, reset, "reset");
		//this->TraceSignals(tf);
	}
	void Update()
	{
		wait(5);
		while(1)
		{
			if(this->iclock.read() == SC_LOGIC_1)
			{
				this->CyclicUpdate();
				wait();
			}
		}
	}
	virtual void TraceSignals(sc_trace_file *tf) = 0;
	virtual void CyclicUpdate() = 0;
	virtual void SequentialUpdate() = 0;
};

#endif // MONITOR_H
