#ifndef ROM_MONITOR_H
#define ROM_MONITOR_H
#include "Monitor.h"
#include "..\Decoder\Components\Component.h"
#define DATA_WIDTH 5
#define SIZE 10
#define ADDRESS_WIDTH 4

#ifndef MTI_SYSTEMC
template<int size = 2, int dataWidth = 2, int addressWidth = 2> class RomMemory : public Component
{
public:
	sc_in<sc_logic> readEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_out< sc_lv<dataWidth> > output;

	sc_signal< sc_lv<dataWidth> > memory[size];
	FILE* file;

	RomMemory(sc_module_name name) : output("output"), readEnable("readEnable"), readAddress("readAddress")
	{
		char* path = ".\\Monitors\\";
		unsigned int length = (unsigned)strlen(path) + (unsigned)strlen(name) + 5;
		char* controlFileName = new char[length];
		strcpy(controlFileName, path);
		strcat(controlFileName, name);
		strcat(controlFileName, ".mif");

#ifdef MTI_SYSTEMC
		file = fopen(controlFileName,"r");
#else
		if (fopen_s(&file, controlFileName,"r") != NULL) 
		{ 
			printf("Failed to open file containing control information.\n"); 
		}
#endif //MTI_SYSTEMC
		fseek(this->file, 0L, SEEK_SET);
		char str[DATA_WIDTH + 1];
		for(int i = 0; i < size; i++)
		{
			fscanf(file,"%s", str);
			sc_lv<dataWidth> temp(str);
			this->memory[i].write(temp);
		}
		fclose(this->file);
		free(controlFileName);
	}
	
	void Update()
	{
		if(this->readEnable.read() == SC_LOGIC_1)
		{
			int address = this->readAddress.read().to_uint();
			this->output.write(this->memory[address].read());
		}
	}
	void Reset()
	{
		this->output.write(0);
	}
};
#else
template <int size = 2, int dataWidth = 2, int addressWidth = 2> class RomMemory : public sc_foreign_module
{
public:
	sc_in<sc_logic> clock;
	sc_in<sc_logic> enable;
	sc_in<sc_logic> reset;
	sc_in<sc_logic> readEnable;
	sc_in< sc_lv<addressWidth> > readAddress;
	sc_out< sc_lv<dataWidth> > output;

    RomMemory(sc_module_name name) : sc_foreign_module(name),
		clock("clock"), enable("enable"), reset("reset"),
		output("output"), readEnable("r_enable"), readAddress("address")
    {
		this->add_parameter("size", size);
		this->add_parameter("width_data", dataWidth);
		this->add_parameter("width_address", addressWidth);

		const char* generic_list[1];
		generic_list[0] = strdup("filename=\"../LDPC/Monitors/rom\"");
		elaborate_foreign_module("work.rom", 1, generic_list);
		free(generic_list);
    }

    ~RomMemory()
    {}
};
#endif //MTI_SYSTEMC

class RomMonitor : public Monitor
{
public:
	RomMemory<SIZE, DATA_WIDTH, ADDRESS_WIDTH>* rom;
	sc_signal<sc_lv<ADDRESS_WIDTH> > readAddress;
	sc_signal<sc_lv<DATA_WIDTH> > output;
	sc_signal<sc_logic> readEnable;

	RomMonitor(sc_module_name name): Monitor(name)
	{
		this->rom = new RomMemory<SIZE, DATA_WIDTH, ADDRESS_WIDTH>("rom");
		this->rom->reset(reset);
		this->rom->clock(iclock);
		this->rom->enable(enable);
		this->rom->readEnable(readEnable);
		this->rom->readAddress(readAddress);
		this->rom->output(output);
	}
	void TraceSignals(sc_trace_file *tf)
	{
	}
	void CyclicUpdate()
	{
	}
	void SequentialUpdate()
	{
		cout << "Otput of ROM:"<<endl;
		this->readEnable.write(SC_LOGIC_1);
		this->readAddress.write(0);
		wait();
		for(int i = 1; i < SIZE; i++)
		{
			this->readAddress.write(i);
			wait();
			cout << output.read().to_uint() << endl;
		}
		wait();
		cout << output.read().to_uint() << endl;
		sc_stop();
	}
};

#endif //ROM_MONITOR_H
