#ifndef INCLUDES_H
#define INCLUDES_H

#define DURATION_PERMUTATION					1
#define DURATION_RAM_ACCESS						1
#define DURATION_ROM_ACCESS						3
#define DURATION_CN								4														// Actual Delay is Plus VariableDelay (cn_degree)
#define DURATION_SN_OUTPUT						DURATION_RAM_ACCESS + 1
#define DURATION_PARTIAL_SUM_CALCULATION		2

#define DELAY_SN								DURATION_ROM_ACCESS
#define DELAY_PERMUTATION_2 					(1 + DURATION_SN_OUTPUT)
#define DELAY_PERMUTATION_1 					(DURATION_CN + DURATION_PERMUTATION)								// Offset to Permutation Network 0 
#define DELAY_PERMUTATION_3 					(DURATION_CN + 3 * DURATION_PERMUTATION + DURATION_PARTIAL_SUM_CALCULATION)		// Offset to Permutation Network 2

#define DELAY_CN								(DELAY_SN + DURATION_SN_OUTPUT + 2 * DURATION_PERMUTATION)
#define DELAY_NEW_SD							(DURATION_SN_OUTPUT + 3 * DURATION_PERMUTATION + DURATION_CN)		// 1 stands for input port of SN
#define DELAY_READ_WRITE						(DURATION_SN_OUTPUT + DURATION_PERMUTATION + 1)

#define WIDTH_BOOL			1
#define WIDTH_INT			32
#define MAX_ITERATION		100
#define WIDTH_ITERATION		9

#include "..\ControlInformation\widths.h"
#ifdef MTI_SYSTEMC
#define CONTROL_PATH "../ControlInformation/"
#else
#define CONTROL_PATH "..\\ControlInformation\\"
#endif

#define SC_INCLUDE_FX
#include <systemc.h>
#include <iostream>
#include <fstream>
#include <assert.h>

// Mathematical Operations on SD and Extrinsic Values
#define WIDTH_SD			6
#define PRECISION_WIDTH		4
#define WIDTH_MAGNITUDE		(WIDTH_SD - 1)


#define SC_LOGIC_TYPE
#ifdef SC_LOGIC_TYPE
typedef sc_lv<WIDTH_SD> TYPE;
#else
typedef int TYPE;
#endif
typedef sc_lv<WIDTH_MAGNITUDE> MAGNITUDE;

static unsigned int MAX_MAGNITUDE;
static int MIN_SD;
static int MAX_SD;
void Calculate_MinMaxValues()
{
	sc_lv<WIDTH_MAGNITUDE> lv;
	for(int i = 0; i < WIDTH_MAGNITUDE; i++)
	{
		lv[i] = SC_LOGIC_1;
	}
	MAX_MAGNITUDE = lv.to_uint();
	MAX_SD = MAX_MAGNITUDE;
	MIN_SD = -MAX_SD; // SD-values are limited to this value, sine magnitude is limited to the same absolute value 
}
inline int CheckLimits(int value)
{
	if(value < MIN_SD)
	{
		value = MIN_SD;
	}
	else if(value > MAX_SD)
	{
		value = MAX_SD;
	}
	return value;
}

#ifdef SC_LOGIC_TYPE
inline TYPE Add_SDs(TYPE operand1, TYPE operand2)
{
	int temp = operand1.to_int() + operand2.to_int();
	return CheckLimits(temp);
}
inline TYPE Subtract_SDs(TYPE operand1, TYPE operand2)
{
	int temp = operand1.to_int() - operand2.to_int();
	return CheckLimits(temp);
}
inline MAGNITUDE Magnitude_of_SD(TYPE operand)
{
	MAGNITUDE temp = (unsigned int)(operand.to_int() < 0 ? (operand.to_int() * (-1)) : operand.to_int());
	return temp;
}
inline bool Sign_of_SD(TYPE operand)
{
	bool temp = operand.to_int() < 0;
	return temp;
}
inline TYPE Magnitude_to_SD(MAGNITUDE operand, bool sign)
{
	int temp = operand.to_uint();
	if(sign)
	{
		temp = -temp;
	}
	return CheckLimits(temp);
}
inline bool Magnitude_Smaller(MAGNITUDE operand1, MAGNITUDE operand2)
{
	return operand1.to_uint() < operand2.to_uint();
}
#else
inline TYPE Add_SDs(TYPE operand1, TYPE operand2)
{
	TYPE temp = operand1 + operand2;
	return temp;
}
inline TYPE Subtract_SDs(TYPE operand1, TYPE operand2)
{
	TYPE temp = operand1 - operand2;
	return temp;
}
inline MAGNITUDE Magnitude_of_SD(TYPE operand)
{
	MAGNITUDE temp = operand < 0 ? (operand * (-1)) : operand;
	return temp;
}
inline bool Sign_of_SD(TYPE operand)
{
	bool temp = operand < 0;
	return temp;
}
inline TYPE Magnitude_to_SD(MAGNITUDE operand, bool sign)
{
	TYPE temp = operand.to_uint();
	if(sign)
	{
		temp = temp * (-1);
	}
	return temp;
}
inline bool Magnitude_Smaller(MAGNITUDE operand1, MAGNITUDE operand2)
{
	return operand1.to_uint() < operand2.to_uint();
}
#endif //SC_LOGIC_TYPE

char* PosIntToString(const char* name, const int number)
{
	int tmp = number;
	int strLength = strlen(name);
	int numLength = 0;
	do{
		tmp /= 10;
		numLength++;
	}while(tmp != 0);

	char* string = (char*) malloc(sizeof(char) * (strLength + numLength + 1));
	if(string == NULL)
		cout << "Not enough memory" << endl;
	strcpy(string, name);
	//string[strLength + numLength] = string[strLength];	// Move Terminating Character
	_itoa(number, &string[strLength], 10);
	#ifndef MTI_SYSTEMC
		_ASSERTE( _CrtCheckMemory( ) );
	#endif // MTI_SYSTEMC
	return string;
}
#endif // INCLUDES_H
