-- Filename     : MemoryDualPort.vhd
-- Author:	    : Andreas Torno
-- Description  : Simultaneous read and write operations supported

library ieee;
USE ieee.std_logic_1164.all;
USE ieee.numeric_std.all;

ENTITY MemoryDualPort IS
	GENERIC (
		size 				: integer := 1;
		width_data 			: integer := 32;
		width_address		: integer := 32
	);
	
    PORT (
		-- Control ports
		clock		: IN std_logic;
		enable		: IN std_logic;
		reset		: IN std_logic;
		r_enable	: IN std_logic;
		w_enable	: IN std_logic;
		r_address	: IN std_logic_vector(width_address - 1 downto 0);
		w_address	: IN std_logic_vector(width_address - 1 downto 0);
		
		-- Data ports
        input     	: IN std_logic_vector(width_data - 1 downto 0);
        output    	: OUT std_logic_vector(width_data - 1 downto 0)
    );
END MemoryDualPort;

ARCHITECTURE RTL OF MemoryDualPort IS
	type memory is array(size - 1 downto 0) of std_logic_vector(width_data - 1 downto 0);
	impure function Initialize return memory is
		variable RAM : memory;
	begin
		for I in memory'range loop
			RAM(i) := (OTHERS => '0');
		end loop;
		return RAM;
	end function;
	signal memory_content : memory := Initialize;
	attribute RAM_STYLE : string;
	attribute RAM_STYLE of memory_content : signal is "DISTRIBUTED"; --{ "DISTRIBUTED" | "BLOCK" };
BEGIN
	process(clock)
    begin
	if clock'event and clock='1' then
		if reset = '1' then
			output <= (others => '0');
	    elsif enable = '1' then
			if w_enable = '1' then
				memory_content(to_integer(unsigned(w_address))) <= input;
			end if;
			if r_enable = '1' then
				output <= memory_content(to_integer(unsigned(r_address)));
			end if;
			if r_enable = '1' and w_enable = '1' and r_address = w_address then
				output <= input;
			end if;
	    end if;
	end if;
    end process;
END RTL;
